import abc
import json
import sqlite3
import uuid
from pydantic import BaseModel
from typing import Iterable, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .recipe import Recipe

PydanticModel = Type[BaseModel]


class Beaker(abc.ABC):
    def __init__(self, name: str, model: PydanticModel, recipe: "Recipe"):
        self.name = name
        self.model = model
        self.recipe = recipe

    def __repr__(self) -> str:
        return f"Beaker({self.name}, {self.model.__name__})"

    @abc.abstractmethod
    def items(self) -> Iterable[tuple[str, BaseModel]]:
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass

    @abc.abstractmethod
    def add_item(self, item: BaseModel, id: str | None = None) -> None:
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        pass

    def add_items(self, items: Iterable[BaseModel]) -> None:
        for item in items:
            self.add_item(item)


class TempBeaker(Beaker):
    def __init__(self, name: str, model: PydanticModel, recipe: "Recipe"):
        super().__init__(name, model, recipe)
        self._items: list[tuple[str, BaseModel]] = []

    def __len__(self) -> int:
        return len(self._items)

    def add_item(self, item: BaseModel, id: str | None = None) -> None:
        if id is None:
            id = str(uuid.uuid1())
        self._items.append((id, item))

    def items(self) -> Iterable[tuple[str, BaseModel]]:
        yield from self._items

    def reset(self) -> None:
        self._items = []


class SqliteBeaker(Beaker):
    def __init__(self, name: str, model: PydanticModel, recipe: "Recipe"):
        super().__init__(name, model, recipe)
        # create table if it doesn't exist
        self.cursor = self.recipe.db.cursor()
        self.cursor.row_factory = sqlite3.Row  # type: ignore
        self.cursor.execute(
            f"CREATE TABLE IF NOT EXISTS {self.name} (uuid TEXT PRIMARY KEY, data JSON)"
        )

    def items(self) -> Iterable[tuple[str, BaseModel]]:
        self.cursor.execute(f"SELECT uuid, data FROM {self.name}")
        data = self.cursor.fetchall()
        for item in data:
            yield item["uuid"], self.model(**json.loads(item["data"]))

    def __len__(self) -> int:
        self.cursor.execute(f"SELECT COUNT(*) FROM {self.name}")
        return self.cursor.fetchone()[0]

    def add_item(self, item: BaseModel, id: str | None = None) -> None:
        if id is None:
            id = str(uuid.uuid1())
            print("UUID", id, item)
        self.cursor.execute(
            f"INSERT INTO {self.name} (uuid, data) VALUES (?, ?)",
            (id, item.model_dump_json()),
        )
        self.recipe.db.commit()

    def reset(self) -> None:
        self.cursor.execute(f"DELETE FROM {self.name}")
        self.recipe.db.commit()
